"""
This file is part of pysofar: A client for interfacing with Sofar Oceans Spotter API

Contents: Abstract implementation Spectrum (see Spectrum1D and Spectrum2D for
implementations)

Copyright (C) 2022
Sofar Ocean Technologies

Authors: Pieter Bart Smit
"""
import numpy
from roguewave.tools import to_datetime, datetime_to_iso_time_string
from typing import TypedDict, List, Tuple, Union
from .windSpotter import U10
from roguewave.wavetheory.lineardispersion import \
    inverse_intrinsic_dispersion_relation, phase_velocity
from datetime import datetime



class WaveSpectrumInput(TypedDict):
    frequency: List[float]
    varianceDensity: List
    timestamp: Union[str,datetime,int,float]
    latitude: float
    longitude: float


class WaveSpectrum():
    """
    Base spectral class.
    """
    frequency_units = 'Hertz'
    angular_units = 'Degrees'
    spectral_density_units = 'm**2/Hertz'
    angular_convention = 'Wave travel direction (going-to), measured anti-clockwise from East'

    def __init__(self,
                 wave_spectrum_input: WaveSpectrumInput
                 ):
        self.frequency = numpy.array(wave_spectrum_input['frequency'])
        self.variance_density = numpy.array(
            wave_spectrum_input['varianceDensity'])
        self.direction = None
        self._a1 = None
        self._b1 = None
        self._a2 = None
        self._b2 = None
        self._e = None
        self.timestamp = to_datetime(wave_spectrum_input['timestamp'])
        self.longitude = wave_spectrum_input['longitude']
        self.latitude = wave_spectrum_input['latitude']

    def frequency_moment(self, power: int, fmin=0, fmax=numpy.inf) -> float:
        pass

    def _create_wave_spectrum_input(self) -> WaveSpectrumInput:
        return WaveSpectrumInput(
            frequency=list(self.frequency),
            varianceDensity=list(self.variance_density),
            timestamp=datetime_to_iso_time_string(self.timestamp),
            latitude=self.latitude,
            longitude=self.longitude
        )

    def _range(self, fmin=0.0, fmax=numpy.inf) -> numpy.ndarray:
        return (self.frequency >= fmin) & (self.frequency < fmax)

    @property
    def radian_direction(self) -> numpy.ndarray:
        return self.direction * numpy.pi / 180

    @property
    def radian_frequency(self) -> numpy.ndarray:
        return self.frequency * 2 * numpy.pi

    @property
    def e(self) -> numpy.array:
        return self._e

    @property
    def a1(self) -> numpy.array:
        return self._a1

    @property
    def b1(self) -> numpy.array:
        return self._b1

    @property
    def a2(self) -> numpy.array:
        return self._a2

    @property
    def b2(self) -> numpy.array:
        return self._b2

    @property
    def A1(self) -> numpy.array:
        return self.a1 * self.e

    @property
    def B1(self) -> numpy.array:
        return self.b1 * self.e

    @property
    def A2(self) -> numpy.array:
        return self.a2 * self.e

    @property
    def B2(self) -> numpy.array:
        return self.b2 * self.e

    def m0(self, fmin=0, fmax=numpy.inf) -> float:
        return self.frequency_moment(0, fmin, fmax)

    def m1(self, fmin=0, fmax=numpy.inf) -> float:
        return self.frequency_moment(1, fmin, fmax)

    def m2(self, fmin=0, fmax=numpy.inf) -> float:
        return self.frequency_moment(2, fmin, fmax)

    def hm0(self, fmin=0, fmax=numpy.inf) -> float:
        return 4 * numpy.sqrt(self.m0(fmin, fmax))

    def tm01(self, fmin=0, fmax=numpy.inf) -> float:
        return self.m0(fmin, fmax) / self.m1(fmin, fmax)

    def tm02(self, fmin=0, fmax=numpy.inf) -> float:
        return numpy.sqrt(self.m0(fmin, fmax) / self.m2(fmin, fmax))

    def peak_index(self, fmin=0, fmax=numpy.inf) -> float:
        range = self._range(fmin, fmax)

        return numpy.argmax(self.e[range])

    def peak_frequency(self, fmin=0, fmax=numpy.inf) -> float:
        return self.frequency[self.peak_index(fmin, fmax)]

    def peak_period(self, fmin=0, fmax=numpy.inf) -> float:
        return 1 / self.peak_frequency(fmin, fmax)

    def peak_direction(self, fmin=0, fmax=numpy.inf):
        index = self.peak_index(fmin, fmax)
        a1 = self.a1[index]
        b1 = self.b1[index]
        return self._mean_direction(a1, b1)

    def peak_spread(self, fmin=0, fmax=numpy.inf):
        index = self.peak_index(fmin, fmax)
        a1 = self.a1[index]
        b1 = self.b1[index]
        return self._spread(a1, b1)

    @staticmethod
    def _mean_direction(a1, b1):
        return numpy.arctan2(b1, a1) * 180 / numpy.pi

    @staticmethod
    def _spread(a1, b1):
        return numpy.sqrt(
            2 - 2 * numpy.sqrt(a1 ** 2 + b1 ** 2)) * 180 / numpy.pi

    @property
    def mean_direction(self):
        return self._mean_direction(self.a1, self.b1)

    @property
    def mean_spread(self):
        return self._spread(self.a1, self.b1)

    def _spectral_weighted(self, property, fmin=0, fmax=numpy.inf):
        range = (self._range(fmin, fmax)) & numpy.isfinite(property)

        return numpy.trapz(property[range] * self.e[range],
                           self.frequency[range]) / self.m0(fmin, fmax)

    def bulk_direction(self, fmin=0, fmax=numpy.inf):
        return self._mean_direction(self.bulk_a1(fmin, fmax),
                                    self.bulk_b1(fmin, fmax))

    def bulk_spread(self, fmin=0, fmax=numpy.inf):
        return self._spread(self.bulk_a1(fmin, fmax), self.bulk_b1(fmin, fmax))

    def bulk_a1(self, fmin=0, fmax=numpy.inf):
        return self._spectral_weighted(self.a1, fmin, fmax)

    def bulk_b1(self, fmin=0, fmax=numpy.inf):
        return self._spectral_weighted(self.b1, fmin, fmax)

    def bulk_a2(self, fmin=0, fmax=numpy.inf):
        return self._spectral_weighted(self.a2, fmin, fmax)

    def bulk_b2(self, fmin=0, fmax=numpy.inf):
        return self._spectral_weighted(self.b2, fmin, fmax)

    def U10(self, **kwargs) -> Tuple[float]:
        windspeed, winddirection, _ = U10(self.e, self.frequency, self.a1,
                                          self.b1, **kwargs)
        return windspeed, winddirection

    def Ustar(self, **kwargs) -> float:
        _, _, _Ustar = U10(self.e, self.frequency, self.a1,
                           self.b1, **kwargs)
        return _Ustar

    def wavenumber(self, depth=numpy.inf, **kwargs):
        return inverse_intrinsic_dispersion_relation(self.radian_frequency,
                                                     depth, **kwargs)

    def peak_wavenumber(self, depth=numpy.inf):
        index = self.peak_index()
        return inverse_intrinsic_dispersion_relation(
            self.radian_frequency[index], depth)

    def peak_wave_age(self, ustar=None, depth=numpy.inf):
        if ustar is None:
            ustar = self.Ustar()
        return phase_velocity(self.peak_wavenumber(depth), depth) / ustar

    def wave_age(self, ustar=None, depth=numpy.inf):
        if ustar is None:
            ustar = self.Ustar()
        return phase_velocity(self.wavenumber(depth), depth) / ustar



"""
This file is part of pysofar: A client for interfacing with Sofar Oceans Spotter API

Contents: 2D Spectrum Class

Copyright (C) 2022
Sofar Ocean Technologies

Authors: Pieter Bart Smit
"""
import numpy
from .wavespectrum import WaveSpectrum, WaveSpectrumInput
from typing import List, Union
from roguewave.tools import datetime_to_iso_time_string


class WaveSpectrum2DInput(WaveSpectrumInput):
    directions: Union[List[float], numpy.ndarray]

class WaveSpectrum2D(WaveSpectrum):
    def __init__(self,
                 wave_spectrum2D_input:WaveSpectrum2DInput
                 ):

        super().__init__(wave_spectrum2D_input)
        self.direction = wave_spectrum2D_input['directions']
        self._a1 = self._directional_moment('a', 1, normalized=True)
        self._b1 = self._directional_moment('b', 1, normalized=True)
        self._a2 = self._directional_moment('a', 2, normalized=True)
        self._b2 = self._directional_moment('b', 2, normalized=True)
        self._e = self._directional_moment('zero', 0, normalized=False)

    def _delta(self):
        angles = self.direction
        forward_diff = (numpy.diff(angles, append=angles[0]) + 180) % 360 - 180
        backward_diff = (numpy.diff(angles,
                                    prepend=angles[-1]) + 180) % 360 - 180
        return (forward_diff + backward_diff) / 2

    def _directional_moment(self, kind='zero', order=0,
                            normalized=True) -> numpy.array:
        delta = self._delta()
        if kind == 'a':
            harmonic = numpy.cos(self.radian_direction * order) * delta
        elif kind == 'b':
            harmonic = numpy.sin(self.radian_direction * order) * delta
        elif kind == 'zero':
            harmonic = delta
        else:
            raise Exception('Unknown moment')
        values = numpy.sum(self.variance_density * harmonic[None, :], axis=-1)

        if normalized:
            scale = numpy.sum(self.variance_density * delta[None, :], axis=-1)
            scale[scale == 0] = 1
        else:
            scale = 1
        return values / scale

    def frequency_moment(self, power: int, fmin=0, fmax=numpy.inf) -> float:
        range = self._range(fmin, fmax)
        return numpy.trapz(self.e[range] * self.frequency[range] ** power,
                           self.frequency[range])

    def _create_wave_spectrum_input(self)->WaveSpectrum2DInput:
        return WaveSpectrum2DInput(
            frequency=list(self.frequency),
            directions=list(self.direction),
            varianceDensity=list(self.variance_density),
            timestamp=datetime_to_iso_time_string(self.timestamp),
            latitude=self.latitude,
            longitude=self.longitude,
        )


from .externaldata.spotter import get_spectrum_from_sofar_spotter_api
from .wavespectra.spectrum1D import WaveSpectrum1D, WaveSpectrum1DInput
from .wavespectra.spectrum2D import WaveSpectrum2D, WaveSpectrum2DInput


import numpy

GRAV = 9.81


def inverse_intrinsic_dispersion_relation(
    angular_frequency,
    dep,
    grav=GRAV,
    maximum_number_of_iterations=10,
    tolerance=1e-3,
):
    """
    Find wavenumber k for a given radial frequency w using Newton Iteration.
    Exit when either maximum number of iterations is reached, or tolerance
    is achieved. Typically only 1 to 2 iterations are needed.

    :param w: radial frequency
    :param dep: depth in meters
    :param grav:  gravitational acceleration
    :param maximum_number_of_iterations: maximum number of iterations
    :param tolerance: relative accuracy
    :return:
    """

    w = numpy.atleast_1d(angular_frequency)

    k_deep_water_estimate = w ** 2 / grav
    k_shallow_water_estimate = w / numpy.sqrt(grav * dep)
    k0 = numpy.zeros(w.shape, dtype=w.dtype)

    # == FIRST GUESS==
    # Use the intersection between shallow and deep water estimates to guestimate
    # which relation to use
    msk = w > numpy.sqrt(grav / dep)
    k0[msk] = k_deep_water_estimate[msk]

    msk = numpy.logical_not(msk)
    k0[msk] = k_shallow_water_estimate[msk]

    # == Newton Iteration ==
    F = numpy.sqrt(k0 * grav * numpy.tanh(k0 * dep)) - w
    cg = numpy.zeros(w.shape, dtype=w.dtype)

    for ii in range(0, maximum_number_of_iterations):
        kd = k0 * dep
        msk = kd > 3
        cg[msk] = 0.5 * w[msk] / k0[msk]
        msk = kd <= 3
        cg[msk] = (1 / 2 + kd[msk] / numpy.sinh(2 * kd[msk])) * w[msk] / k0[msk]
        # cg = (1 / 2 + kd / numpy.sinh(2 * kd)) * w / k0
        k0 = k0 - F / cg

        F = numpy.sqrt(k0 * grav * numpy.tanh(k0 * dep)) - w
        error = numpy.abs(F) / w
        if numpy.all(error < tolerance):
            break

    return k0


def intrinsic_dispersion_relation(k, dep, grav=GRAV):
    """
    :param k: Wavenumber (rad/m)
    :param depth: Depth (m)
    :param grav: Gravitational acceleration (m/s^2)
    :return:
    """
    k = numpy.atleast_1d(k)
    w = numpy.sqrt(grav * k * numpy.tanh(k * dep))
    return w


def phase_velocity(k, depth, grav=GRAV):
    """
    :param k: Wavenumber (rad/m)
    :param depth: Depth (m)
    :param grav: Gravitational acceleration (m/s^2)
    :return:
    """
    return intrinsic_dispersion_relation(k, depth, grav=GRAV) / k


def ratio_group_velocity_to_phase_velocity(k, depth, grav):
    """
    :param k: Wavenumber (rad/m)
    :param depth: Depth (m)
    :param grav: Gravitational acceleration (m/s^2)
    :return:
    """
    kd = k * depth
    n = numpy.zeros(kd.shape, dtype=k.dtype)

    msk = kd > 3
    n[msk] = 0.5

    msk = kd <= 3
    n[msk] = 0.5 + kd[msk] / numpy.sinh(2 * kd[msk])

    return n


def intrinsic_group_velocity(k, depth, grav=GRAV):
    """
    :param k: Wavenumber (rad/m)
    :param depth: Depth (m)
    :param grav: Gravitational acceleration (m/s^2)
    :return:
    """
    return ratio_group_velocity_to_phase_velocity(k, depth, grav=GRAV) * phase_velocity(k, depth, grav)


def jacobian_wavenumber_to_radial_frequency(k, depth, grav=GRAV):
    """
    :param k: Wavenumber (rad/m)
    :param depth: Depth (m)
    :param grav: Gravitational acceleration (m/s^2)
    :return:
    """
    # return numpy.ones( k.shape  )
    return 1 / intrinsic_group_velocity(k, depth, grav)


def jacobian_radial_frequency_to_wavenumber(k, depth, grav=GRAV):
    """
    :param k: Wavenumber (rad/m)
    :param depth: Depth (m)
    :param grav: Gravitational acceleration (m/s^2)
    :return:
    """
    return intrinsic_group_velocity(k, depth, grav)


# Aliasses based on common notation in linear wave theory
c = phase_velocity
cg = intrinsic_group_velocity
k = inverse_intrinsic_dispersion_relation
w = intrinsic_dispersion_relation
n = ratio_group_velocity_to_phase_velocity



"""
This file is part of pysofar: A client for interfacing with Sofar Oceans Spotter API

Contents: Spectral estimator

Copyright (C) 2022
Sofar Ocean Technologies

Authors: Pieter Bart Smit
"""

import numpy

def mem(directions_radians: numpy.ndarray, a1, b1, a2, b2) -> numpy.ndarray:
    """
    This function uses the maximum entropy method by Lygre and Krogstadt (1986,JPO)
    to estimate the directional shape of the spectrum.

    Lygre, A., & Krogstad, H. E. (1986). Maximum entropy estimation of the directional
    distribution in ocean wave spectra. Journal of Physical Oceanography, 16(12), 2052-2060.

    :param direction_radians: radian directions (going to, anti-clockswise from east) want to
                              evaluate the spectrum on
    :param a1: 1st cosine Fourier coefficient of the directional distribution
    :param b1: 1st sine Fourier coefficient of the directional distribution
    :param a2: 2nd cosine Fourier coefficient of the directional distribution
    :param b2: 2nd sine Fourier coefficient of the directional distribution

    :return: Directional distribution as a numpy array

    Note that:
    d1 = a1; d2 =b1; d3 = a2 and d4=b2 in the defining equations 10.
    """

    # Ensure that these are numpy arrays
    a1 = numpy.atleast_1d(a1)
    b1 = numpy.atleast_1d(b1)
    a2 = numpy.atleast_1d(a2)
    b2 = numpy.atleast_1d(b2)

    number_of_directions = len(directions_radians)

    c1 = a1 + 1j * b1
    c2 = a2 + 1j * b2
    #
    # Eq. 13 L&K86
    #
    Phi1 = (c1 - c2 * numpy.conj(c1)) / (1 - c1 * numpy.conj(c1))
    Phi2 = c2 - Phi1 * c1
    #
    e1 = numpy.exp(-directions_radians * 1j)
    e2 = numpy.exp(-directions_radians * 2j)

    numerator = (1 - Phi1 * numpy.conj(c1) - Phi2 * numpy.conj(c2))
    denominator = numpy.abs(1 - Phi1[:,None] * e1[None,:]
                              - Phi2[:,None] * e2[None,:]) ** 2

    D = numpy.real( numerator[:,None] / denominator  ) / numpy.pi / 2

    # Normalize to 1. in discrete sense
    integralApprox = numpy.sum(D,axis=-1) * numpy.pi * 2. / number_of_directions
    D = D / integralApprox[:,None]

    return numpy.squeeze(D)


"""

Copyright (C) 2022
Sofar Ocean Technologies

Authors: Pieter Smit
"""
import os
import dotenv
import requests
import json
from pysofar import SofarConnection
from pysofar.wavefleet_exceptions import QueryError
from functools import cached_property
from roguewave.wavespectra.spectrum2D import WaveSpectrum2D, \
    WaveSpectrum2DInput
import typing
import netCDF4
import numpy


def get_token():
    # config values
    userpath = os.path.expanduser("~")
    enviromentFile = os.path.join(userpath, 'sofar_api.env')
    dotenv.load_dotenv(enviromentFile)
    token = os.getenv('SPECTRAL_API_TOKEN')
    _wavefleet_token = token

    return _wavefleet_token


def get_endpoint():
    _endpoint = os.getenv('SPECTRAL_URL')
    if _endpoint is None:
        _endpoint = 'https://api.sofarocean.com/api'
    return _endpoint


class SofarSpectralAPI(SofarConnection):
    def __init__(self, costum_token=None):
        #
        token = costum_token if costum_token else get_token()
        super().__init__(token)
        self.endpoint = get_endpoint()

    @cached_property
    def _query(self):
        scode, data = self._get('op-wave-spectra')

        if scode != 200:
            raise QueryError(data['message'])

        return data['data']

    def _download(self, latitude, longitude):
        # Helper methods to download files

        url = f"{self.endpoint}/op-wave-spectra/{latitude}/{longitude}"
        response = requests.get(url, headers=self.header)

        status = response.status_code
        if status != 200:
            data = json.loads(response.text)
            raise QueryError(data['message'])

        data = response.content
        return data

    def points(self):
        data = self._query
        print(data)
        return [{'latitude': x['latitude'], 'longitude': x['longitude']} for x
                in data]

    def urls(self):
        data = self._query
        return [x['url'] for x in data]

    def download_spectral_file(self, latitude, longitude, directory) -> str:
        os.makedirs(directory, exist_ok=True)
        filename = f'spectrum_{longitude}E_{latitude}N.nc'
        filepath = os.path.join(directory, filename)

        with open(filepath, 'wb') as file:
            file.write(self._download(latitude, longitude))

        return filepath

    def download_all_spectral_files(self, directory) -> typing.List[str]:
        points = self.points()
        return [
            self.download_spectral_file(point['latitude'], point['longitude'],
                                        directory) for point in points]

def load_spectral_file(file)->typing.List[WaveSpectrum2D]:
    dataset = netCDF4.Dataset(file)

    nt = dataset.dimensions['time'].size
    frequencies = dataset.variables['frequencies'][:]
    directions = dataset.variables['directions'][:]

    spectra = []
    for ii in range(0,nt):
        unixepochtime = dataset.variables['time'][ii]
        latitude = dataset.variables['latitude'][0]
        longitude = dataset.variables['longitude'][0]
        variance_density = \
            numpy.squeeze(dataset.variables['frequency_direction_spectrum'][ii,:,:])

        spectrum2D_input = WaveSpectrum2DInput(
            frequency=frequencies,varianceDensity=variance_density,
            timestamp=unixepochtime,latitude=latitude,longitude=longitude,
            directions=directions
        )
        spectra.append(WaveSpectrum2D(spectrum2D_input))
    return spectra


from roguewave.wavespectra.spectrum1D import WaveSpectrum1D, \
    WaveSpectrum1DInput
import typing
from pysofar.spotter import Spotter
from datetime import datetime
from roguewave.tools import datetime_to_iso_time_string
from .exceptions import ExceptionNoFrequencyData

def get_spectrum_from_sofar_spotter_api(
        spotter: Spotter,
        start_date: typing.Union[datetime, str] = None,
        end_date: typing.Union[datetime, str] = None,
        limit: int = 20,
) -> typing.List[WaveSpectrum1D]:
    """
    Grabs the requested spectra for this spotter based on the given keyword arguments

    :param limit: The limit for data to grab. Defaults to 20, For frequency data max of 100 samples at a time.
    :param start_date: ISO 8601 formatted date string. If not included defaults to beginning of spotters history
    :param end_date: ISO 8601 formatted date string. If not included defaults to end of spotter history

    :return: Data as a FrequencyDataList Object
    """

    start_date = datetime_to_iso_time_string(start_date)
    end_date = datetime_to_iso_time_string(end_date)

    json_data = spotter.grab_data(
        limit=limit,
        start_date=datetime_to_iso_time_string(start_date),
        end_date=datetime_to_iso_time_string(end_date),
        include_frequency_data=True,
        include_directional_moments=True)

    if not json_data['frequencyData']:
        raise ExceptionNoFrequencyData(
            f'Spotter {spotter.id} has no spectral data for the requested time range')

    out = []
    for spectrum in json_data['frequencyData']:
        # Load the data into the input object- this is a one-to-one
        # mapping of keys between dictionaries (we would not _need_ to
        # map onto a input object-> we could pass spectrum directly to
        # the constructor. But it is clearer to fail here if something
        # is amiss with the api JSON response format)
        wave_spectrum_input = WaveSpectrum1DInput(**spectrum)

        out.append(WaveSpectrum1D(wave_spectrum_input))

    return out



"""
This file is part of pysofar: A client for interfacing with Sofar Oceans Spotter API

Contents: 1D Spectrum Class

Copyright (C) 2022
Sofar Ocean Technologies

Authors: Pieter Bart Smit
"""

import numpy
from .spectrum2D import WaveSpectrum2D, WaveSpectrum2DInput
from .wavespectrum import WaveSpectrum, WaveSpectrumInput
from roguewave.tools import datetime_to_iso_time_string
from .mem import mem
from typing import List, Union

class WaveSpectrum1DInput(WaveSpectrumInput):
    a1: Union[List[float], numpy.ndarray]
    b1: Union[List[float], numpy.ndarray]
    a2: Union[List[float], numpy.ndarray]
    b2: Union[List[float], numpy.ndarray]

class WaveSpectrum1D(WaveSpectrum):
    spectral_density_units = 'm**2/Hertz'

    def __init__(self,
                 wave_spectrum1D_input:WaveSpectrum1DInput
                 ):
        super().__init__(wave_spectrum1D_input)
        self._a1 = numpy.array(wave_spectrum1D_input['a1'])
        self._b1 = numpy.array(wave_spectrum1D_input['b1'])
        self._b2 = numpy.array(wave_spectrum1D_input['b2'])
        self._a2 = numpy.array(wave_spectrum1D_input['a2'])
        self._e = numpy.array(wave_spectrum1D_input['varianceDensity'])

    def frequency_moment(self, power: int, fmin=0, fmax=numpy.inf) -> float:
        range = self._range(fmin,fmax)

        return numpy.trapz(
            self.variance_density[range] * self.frequency[range] ** power,
            self.frequency[range])

    def _create_wave_spectrum_input(self)->WaveSpectrum1DInput:
        return WaveSpectrum1DInput(
            frequency=list(self.frequency),
            varianceDensity=list(self.variance_density),
            timestamp=datetime_to_iso_time_string(self.timestamp),
            latitude=self.latitude,
            longitude=self.longitude,
            a1=list(self.a1),
            b1=list(self.b1),
            a2=list(self.a2),
            b2=list(self.b2)
            )

    def spectrum2d(self, number_of_directions: int,
                   method:str='maximum_entropy_method')->WaveSpectrum2D:
        """
        Construct a 2D spectrum based on the 1.5D spectrum and a spectral
        reconstruction method.

        :param number_of_directions: length of the directional vector for the
        2D spectrum. Directions returned are in degrees
        """
        direction = numpy.linspace(0, 360, number_of_directions,
                                   endpoint=False)

        # Jacobian to transform distribution as function of radian angles into
        # degrees.
        Jacobian = numpy.pi / 180

        if method.lower() in ['maximum_entropy_method', 'mem']:
            # reconstruct the directional distribution using the maximum entropy
            # method.
            D = mem(direction * numpy.pi / 180, self.a1, self.b1, self.a2,
                    self.b2) * Jacobian
        else:
            raise Exception(f'unsupported spectral estimator method: {method}')

        wave_spectrum2D_input = WaveSpectrum2DInput(
            frequency=self.frequency,
            directions=direction,
            varianceDensity=self.variance_density[:, None] * D,
            timestamp=self.timestamp,
            longitude=self.longitude,
            latitude=self.latitude
        )

        # We return a 2D wave spectrum object.
        return WaveSpectrum2D(wave_spectrum2D_input)


